import argparse
import itertools
import logging
import multiprocessing
import os
import errno
import subprocess
import time
from datetime import datetime
from pathlib import PurePath

import numpy as np
from dask_jobqueue import SLURMCluster
from distributed import Client
from google.cloud.storage import Blob
from aind_data_transfer.gcs import create_client, GCSUploader
from aind_data_transfer.util.file_utils import collect_filepaths, make_cloud_paths

from cluster.config import load_jobqueue_config

logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M")
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


GCLOUD_TOOLS = ['gcloud alpha storage', 'gsutil']


def get_client(deployment="slurm"):
    """
    Args:
        deployment (str): type of cluster
    Returns:
        the dask Client and its configuration dict
    """
    base_config = load_jobqueue_config()
    if deployment == "slurm":
        config = base_config["jobqueue"]["slurm"]
        # cluster config is automatically populated from
        # ~/.config/dask/jobqueue.yaml
        cluster = SLURMCluster()
        cluster.scale(config["n_workers"])
    else:
        raise NotImplementedError
    logger.info(cluster.job_script())
    client = Client(cluster)
    return client, config


def _chunk_files(filepaths, n_workers, tasks_per_worker):
    """
    Partition a list of filepaths into chunks.
    Our target is ~tasks_per_worker chunks of files for each dask worker to upload.
    The way each job (upload of a single chunk) is actually distributed is up to the scheduler.
    Some workers may get many jobs while others get few.
    Increasing tasks_per_worker can help the scheduler balance jobs among workers.
    Args:
        filepaths (list): list of absolute paths to chunk
        n_workers (int): the number of dask workers
        tasks_per_worker (int): target # tasks per dask worker
    Returns:
        a list of lists, where each sublist contains a list of filepaths
    """
    n = min(n_workers * tasks_per_worker, len(filepaths))
    chunked_files = np.array_split(filepaths, n)
    logger.info(
        f"Split files into {len(chunked_files)} chunks with "
        f"{len(chunked_files[0])} files each"
    )
    return chunked_files


def _python_upload_worker(
    bucket_name, files, gcs_path, root=None, chunk_size=256 * 1024 * 1024
):
    """
    Uploads a list of filepaths to GCS using the google-cloud-storage Python API
    Args:
        bucket_name (str): name of bucket
        files (list): list of absolute paths to upload to bucket
        gcs_path (str): google cloud storage path to upload files to
        root (str): directory shared by files to serve as root directory
        chunk_size (int): set the blob chunk size (bytes).
                          Increasing this can speed up transfers.
    """
    uploader = GCSUploader(bucket_name)
    return uploader.upload_files(files, gcs_path, root, chunk_size=chunk_size)


def run_cluster_job(
    filepaths,
    bucket,
    gcs_path,
    tasks_per_worker,
    root=None,
    chunk_size=256 * 1024 * 1024,
):
    """
    Upload a list of filepaths using multiple distributed instances of the google-cloud-storage Python API
    Args:
        filepaths (list): list of absolute paths to upload
        bucket (str): name of the bucket
        gcs_path (str): cloud storage location to store uploaded files
        tasks_per_worker (int): target number of chunks for each worker to process
        root (str): shared directory for filepaths to serve as root folder in the bucket
        chunk_size (int): set the blob chunk size (bytes).
                          Increasing this can speed up transfers.
    """
    client, config = get_client()
    ntasks = config["n_workers"]

    chunked_files = _chunk_files(filepaths, ntasks, tasks_per_worker)

    futures = []
    for i, chunk in enumerate(chunked_files):
        futures.append(
            client.submit(
                _python_upload_worker,
                bucket_name=bucket,
                files=chunk,
                gcs_path=gcs_path,
                root=root,
                chunk_size=chunk_size,
            )
        )
    failed_uploads = list(itertools.chain(*client.gather(futures)))
    logger.info(f"{len(failed_uploads)} failed uploads:\n{failed_uploads}")
    client.shutdown()


def run_python_local_job(
    input_dir, bucket, gcs_path, n_workers=4, chunk_size=256 * 1024 * 1024, exclude_dirs=None
):
    """
    Upload a directory using the google-cloud-storage Python API and multiprocessing
    Args:
        input_dir (str): directory to upload
        bucket (str): name of the bucket
        gcs_path (str): cloud storage location to store uploaded files
        n_workers (int): number of workers
        chunk_size (int): set the blob chunk size (bytes).
                          Increasing this can speed up transfers.
        exclude_dirs (list): list of directory names to exclude, e.g., ["dir1", "dir2"]
    """
    files = collect_filepaths(input_dir, recursive=True, exclude_dirs=exclude_dirs)
    chunked_files = _chunk_files(files, n_workers, tasks_per_worker=1)
    args = zip(
        itertools.repeat(bucket),
        chunked_files,
        itertools.repeat(gcs_path),
        itertools.repeat(input_dir),
        itertools.repeat(chunk_size),
    )
    with multiprocessing.Pool(n_workers) as pool:
        failed_uploads = list(
            itertools.chain(*pool.starmap(_python_upload_worker, args))
        )
    logger.info(f"{len(failed_uploads)} failed uploads:\n{failed_uploads}")


def _parse_cp_failures(log_path):
    """
    Parse the gsutil cp log for any transfers with a non-OK status code
    Args:
        log_path (str): path to the log file created by gsutil,
                        e.g., gsutil cp -L log_path.log ...
    Returns:
        a list of filepaths that resulted in an error during transfer
    """
    failures = []
    with open(log_path, "r") as f:
        next(f)  # skip header
        for line in f:
            parts = line.strip().split(",")
            src = parts[0]
            status = parts[-2]
            if status != "OK":
                failures.append(src)
    return failures


def _make_boto_options(n_threads):
    """Create Boto options to pass to gsutil
    See https://cloud.google.com/storage/docs/boto-gsutil
    Args:
        n_threads (int): number of threads to use for parallel uploads
    Returns:
        a list of option strings
    """
    options = []
    # Use the same process / thread ratio as the ~/.boto defaults
    # These defaults are different on Windows (1 process, many threads)
    processes = max(1, n_threads - 1)
    options.append(f"Boto:parallel_thread_count={n_threads}")
    options.append(f"Boto:parallel_process_count={processes}")
    return options


def build_gcloud_cmd(tool, func, input_dir, bucket, gcs_path, nthreads, recursive=True, logfile=None):
    if tool not in GCLOUD_TOOLS:
        raise ValueError(f"Invalid tool {tool}")
    cmd = f"{tool}"
    if tool == "gsutil" and nthreads > 1:
        sep = " -o "
        cmd += f" -m {sep}{sep.join(_make_boto_options(nthreads))}"
    cmd += f" {func}"
    if logfile is not None:
        cmd += f" -L {logfile}"
    if recursive:
        cmd += " -r"
    cmd += f" {input_dir} gs://{bucket}/{gcs_path}"
    logger.debug(f"Built command: {cmd}")
    return cmd


def run_gcloud_local_job(tool, input_dir, bucket, gcs_path, nthreads):
    """
    Upload a directory using the gcloud storage CLI
    Args:
        tool (str): either "gcloud alpha storage" or "gsutil"
        input_dir (str): directory to upload
        bucket (str): name of the bucket
        gcs_path (str): cloud storage location to store uploaded files
        nthreads (int): num threads to use for upload. Only applies to gsutil
    """
    # assumes there will only be a single space, which should always be the case
    logfile = f"{tool.replace(' ', '_')}-log-{datetime.utcnow().strftime('%Y%m%d%H%M%S%f')}.log"
    cmd = build_gcloud_cmd(
        tool,
        func="cp",
        input_dir=input_dir,
        bucket=bucket,
        gcs_path=gcs_path,
        nthreads=nthreads,
        logfile=logfile
    )
    ret = subprocess.run(cmd, shell=True)
    if ret.returncode != 0:
        logger.error(f"{tool} exited with code {ret.returncode}")
        failed_uploads = _parse_cp_failures(logfile)
        logger.error(f"{len(failed_uploads)} failed uploads:\n{failed_uploads}")
    try:
        os.remove(logfile)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def validate_blobs(bucket_name, target_paths):
    """
    Ensure that all blobs were uploaded correctly
    Args:
        bucket_name (str): bucket name
        target_paths (list): list of cloud paths to validate
    Returns:
        list of cloud storage paths that are missing a Blob
    """
    storage_client = create_client()
    bucket = storage_client.get_bucket(bucket_name)
    missing_paths = []
    for path in target_paths:
        if not Blob(path, bucket).exists():
            missing_paths.append(path)
    return missing_paths


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        help="folder or file to upload",
    )
    parser.add_argument(
        "-b",
        "--bucket",
        type=str,
        help="s3 bucket",
    )
    parser.add_argument(
        "--gcs_path",
        type=str,
        default=None,
        help='gcs path relative to bucket, e.g., "folder/data.h5"',
    )
    parser.add_argument(
        "--recursive",
        default=False,
        action="store_true",
        help="upload a folder recursively",
    )
    parser.add_argument(
        "--tasks_per_worker",
        type=int,
        default=3,
        help="Split the work so that each worker has this many jobs. "
        "Increasing this can help with worker/task imbalance.",
    )
    parser.add_argument(
        "--cluster",
        default=False,
        action="store_true",
        help="run in cluster mode",
    )
    parser.add_argument(
        "--nthreads",
        type=int,
        default=1,
        help="number of threads if running locally",
    )
    parser.add_argument(
        "--chunk_size",
        type=int,
        default=256,
        help="upload files in chunks of this size (MB)",
    )
    parser.add_argument(
        "--method",
        choices=['python', 'gsutil', 'gcloud_alpha_storage'],
        default="python",
        help="use either gsutil, gcloud storage or the google-cloud-storage Python API for local upload."
             "This does not apply when --cluster is used.",
    )
    parser.add_argument(
        "--validate",
        default=False,
        action="store_true",
        help="Validate that all uploads were successful. This can take a while if there are many files."
        "This is an additional billed API request for each uploaded object.",
    )
    parser.add_argument(
        "--exclude_dirs",
        type=str,
        nargs="+",
        default=None,
        help="directories to exclude from upload"
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    gcs_path = args.gcs_path
    if gcs_path is None:
        gcs_path = PurePath(args.input).name
    # remove leading slash since it will result in the structure
    # "bucket/'/'/s3_path" instead of "bucket/s3_path"
    gcs_path = gcs_path.strip("/")
    logger.info(f"Will upload to {args.bucket}/{gcs_path}")

    filepaths = collect_filepaths(args.input, recursive=args.recursive, exclude_dirs=args.exclude_dirs)
    cloud_paths = make_cloud_paths(filepaths, args.gcs_path, args.input)

    t0 = time.time()
    if args.cluster:
        chunk_size = args.chunk_size * 1024 * 1024  # MB to bytes
        run_cluster_job(
            filepaths=filepaths,
            bucket=args.bucket,
            gcs_path=args.gcs_path,
            tasks_per_worker=args.tasks_per_worker,
            root=args.input,
            chunk_size=chunk_size,
        )
    else:
        # replace CLI friendly string (no spaces) with actual gcloud command string
        norm_method = args.method.replace("_", " ")
        if norm_method in GCLOUD_TOOLS:
            # TODO: add exclusions
            run_gcloud_local_job(
                norm_method, args.input, args.bucket, args.gcs_path, args.nthreads
            )
        elif norm_method == "python":
            run_python_local_job(
                args.input, args.bucket, args.gcs_path, args.nthreads, exclude_dirs=args.exclude_dirs
            )
        else:
            raise ValueError(f"Unsupported method {norm_method}")

    logger.info(f"Upload done. Took {time.time() - t0}s ")

    if args.validate:
        missing_blobs = validate_blobs(
            bucket_name=args.bucket, target_paths=cloud_paths
        )
        if not missing_blobs:
            logger.info("All uploads were successful")
        else:
            logger.info(
                f"{len(missing_blobs)} blobs missing.\n{missing_blobs}"
            )


if __name__ == "__main__":
    main()



import os
import unittest
from pathlib import Path

import numpy as np
from numcodecs import Blosc
from wavpack_numcodecs import WavPack

from aind_data_transfer.readers import EphysReaders
from aind_data_transfer.transformations.compressors import (
    EphysCompressors,
    ImagingCompressors,
    VideoCompressor,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
RESOURCES_DIR = TEST_DIR / "resources"


class TestEphysCompressors(unittest.TestCase):
    def test_get_compressor(self):
        blosc_configs = {
            "cname": "zstd",
            "clevel": 9,
            "shuffle": Blosc.BITSHUFFLE,
        }
        wavpack_configs = {"level": 3}
        blosc = EphysCompressors.get_compressor(
            EphysCompressors.Compressors.blosc.name, **blosc_configs
        )
        wavpack = EphysCompressors.get_compressor(
            EphysCompressors.Compressors.wavpack.name, **wavpack_configs
        )
        expected_blosc = Blosc(
            cname="zstd", clevel=9, shuffle=Blosc.BITSHUFFLE
        )
        expected_wavpack = WavPack(level=3)
        self.assertEqual(blosc, expected_blosc)
        self.assertEqual(wavpack, expected_wavpack)

    def test_get_compressor_fails(self):
        with self.assertRaises(Exception):
            EphysCompressors.get_compressor("Made up name")

    def test_scale_recording(self):
        open_ephys_dir = (
            RESOURCES_DIR / "v0.6.x_neuropixels_multiexp_multistream"
        )
        openephys_reader = EphysReaders.Readers.openephys.name
        read_blocks = EphysReaders.get_read_blocks(
            openephys_reader, open_ephys_dir
        )
        read_block = next(read_blocks)
        chunk_size = min(read_block["recording"].get_num_frames(0) - 1, 10000)
        # The first read block should be a NI-DAQ recording.
        self.assertTrue(
            EphysReaders.RecordingBlockPrefixes.nidaq.value
            in read_block["stream_name"],
            "The first block isn't NI-DAQ. If the test data has "
            "been updated, this test needs to be updated too.",
        )
        # We expect the NI-DAQ scaled recordings to be the same as the original
        scaled_read_blocks = EphysCompressors.scale_read_blocks(
            [read_block], disable_tqdm=True, chunk_size=chunk_size
        )
        scaled_read_block = next(scaled_read_blocks)
        self.assertEqual(
            scaled_read_block["scaled_recording"], read_block["recording"]
        )

        read_block = next(read_blocks)
        # The second read block should be a Neuropixel recording.
        self.assertTrue(
            EphysReaders.RecordingBlockPrefixes.neuropix.value
            in read_block["stream_name"],
            "The second block isn't Neuropixel. If the test data "
            "has been updated, this test needs to be updated too.",
        )
        chunk_size = min(read_block["recording"].get_num_frames(0) - 1, 10000)
        scaled_read_blocks = EphysCompressors.scale_read_blocks(
            [read_block], disable_tqdm=True, chunk_size=chunk_size
        )
        scaled_read_block = next(scaled_read_blocks)

        # The lsb value should be 1 and the median values should be 0 in the
        # scaled neuropixel recordings.
        lsb_value, median_values = EphysCompressors._get_median_and_lsb(
            scaled_read_block["scaled_recording"],
            disable_tqdm=True,
            num_chunks_per_segment=10,
            chunk_size=chunk_size,
        )

        expected_median_values = np.zeros(
            shape=median_values.shape, dtype=median_values.dtype
        )

        self.assertEqual(lsb_value, 1)
        self.assertTrue(np.array_equal(expected_median_values, median_values))


class TestImagingCompressors(unittest.TestCase):
    def test_get_compressor(self):
        blosc_configs = {
            "cname": "zstd",
            "clevel": 1,
            "shuffle": Blosc.SHUFFLE,
        }
        blosc = ImagingCompressors.get_compressor(
            ImagingCompressors.Compressors.blosc.name, **blosc_configs
        )
        expected_blosc = Blosc(cname="zstd", clevel=1, shuffle=Blosc.SHUFFLE)
        self.assertEqual(blosc, expected_blosc)

    def test_get_compressor_fails(self):
        with self.assertRaises(Exception):
            ImagingCompressors.get_compressor("Made up name")


class TestVideoCompressor(unittest.TestCase):
    """Tests VideoCompressor class methods"""

    def test_video_compressor_creation(self):
        """Tests VideoCompressor default is set correctly"""
        video_compressor = VideoCompressor()
        self.assertEqual(video_compressor.compression_level, 5)


if __name__ == "__main__":
    unittest.main()



import fnmatch
import os
from pathlib import PurePath, PurePosixPath, Path
from typing import List, Optional, Tuple, Union

from aind_data_transfer.util.io_utils import DataReaderFactory


def collect_filepaths(
    folder: Union[str, os.PathLike],
    recursive: bool = True,
    include_exts: Optional[List[str]] = None,
    exclude_dirs: Optional[List[str]] = None,
) -> List[str]:
    """Get the absolute paths for all files in folder
    Args:
        folder (str): the directory to look for files
        recursive (bool): whether to traverse all sub-folders
        include_exts (optional): list of valid file extensions to include.
                                 e.g., ['.tiff', '.h5', '.ims']
        exclude_dirs (optional): list of directories to exclude from the search
    Returns:
        list of filepaths
    """
    if exclude_dirs is None:
        exclude_dirs = []
    filepaths = []
    for root, _, files in os.walk(folder):
        root_name = Path(root).name
        if root_name in exclude_dirs:
            continue
        for f in files:
            path = os.path.join(root, f)
            _, ext = os.path.splitext(path)
            if include_exts is None or ext in include_exts:
                filepaths.append(path)
        if not recursive:
            break
    return filepaths


def get_images(
        image_folder: Union[str, os.PathLike],
        exclude: List[str] = None,
        include_exts: List[str] = DataReaderFactory().VALID_EXTENSIONS,
        recursive: bool = False
) -> List[str]:
    """Get the absolute paths for all images in a folder
    Args:
        image_folder: the directory to look for images
        exclude: list of filename patterns to exclude
        include_exts: list of valid file extensions to include.
                                 e.g., ['.tiff', '.h5', '.ims']
        recursive: whether to traverse all sub-folders
    Returns:
        list of image paths
    """
    if exclude is None:
        exclude = []
    image_paths = collect_filepaths(
        image_folder,
        recursive=recursive,
        include_exts=include_exts,
    )

    exclude_paths = set()
    for path in image_paths:
        if any(fnmatch.fnmatch(path, pattern) for pattern in exclude):
            exclude_paths.add(path)

    image_paths = [p for p in image_paths if p not in exclude_paths]

    return image_paths


def join_cloud_paths(cloud_dest_path: str, relpath: str) -> str:
    """Always produce posix-style paths, even if relpath
    is Windows-style
    Args:
        cloud_dest_path (str): first part of the path
        relpath (str): second part of the path
    Returns:
        the joined path
    """
    cloud_dest_path = PurePosixPath(cloud_dest_path)
    relpath = PurePath(relpath)
    return str(cloud_dest_path / relpath)


def make_cloud_paths(
    filepaths: List[Union[str, os.PathLike]],
    cloud_dest_path: Union[str, os.PathLike],
    root: Union[str, os.PathLike] = None,
) -> List[str]:
    """
    Given a list of filepaths and a cloud destination folder,
    build a cloud path for each file relative to root.
    Args:
        filepaths (list): list of paths
        cloud_dest_path (str): the cloud storage path to store files
        root (str): a directory shared by all paths in filepaths, which will
                    serve as the new root under cloud_dest path. If none,
                    all files are uploaded as a flat list to cloud_dest_path,
                    ignoring any exiting directory structure.
    Returns:
        list of cloud storage paths
    Examples:
    >>> filepaths = ["/data/micr/0001.tif", "/data/metadata/rig.json"]
    >>> root = "/data"
    >>> cloud_dest_path = "my-data"
    >>> cloud_paths = make_cloud_paths(filepaths, cloud_dest_path, root)
    >>> print(cloud_paths)
    ['my-data/micr/001.tif', 'my-data/metadata/rig.json']
    >>> cloud_paths = make_cloud_paths(filepaths, cloud_dest_path, None)
    >>> print(cloud_paths)
    ['my-data/001.tif', 'my-data/rig.json']
    """
    cloud_paths = []
    # remove both leading and trailing '/'
    cloud_dest_path = cloud_dest_path.strip("/")
    for fpath in filepaths:
        if root is None:
            cloud_paths.append(
                join_cloud_paths(cloud_dest_path, PurePath(fpath).name)
            )
        else:
            cloud_paths.append(
                join_cloud_paths(cloud_dest_path, os.path.relpath(fpath, root))
            )
    return cloud_paths


def is_cloud_url(url: str):
    """
    Test if the url points to an AWS S3 or Google Cloud Storage URI
    Args:
        url: the url to test
    Returns:
        True if url is a cloud url
    """
    url = str(url)
    if url.startswith("s3://"):
        return True
    if url.startswith("gs://"):
        return True
    return False


def parse_cloud_url(cloud_url: str) -> Tuple[str, str, str]:
    """
    Get the cloud storage provider, bucket name, and path
    from an AWS S3 or Google Cloud Storage url.
    Args:
        cloud_url: the cloud url to parse
    Returns:
        a tuple containing the provider, bucket and path
    """
    parts = Path(cloud_url).parts
    provider = parts[0] + "//"
    bucket = parts[1]
    cloud_dst = "/".join(parts[2:])
    return provider, bucket, cloud_dst



import os

import yaml
from yaml import Loader

_HOME = os.getenv("HOME")
# This is where dask-jobqueue looks for the cluster configuration
DASK_CONF_FILE = f"{_HOME}/.config/dask/jobqueue.yaml"


def load_jobqueue_config():
    with open(DASK_CONF_FILE, "r") as f:
        config = yaml.load(f, Loader)
    return config



import os
import unittest
from pathlib import Path

from aind_data_transfer.readers import EphysReaders, ImagingReaders
from parameterized import parameterized

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
RESOURCES_DIR = TEST_DIR / "resources"


class TestEphysReaders(unittest.TestCase):

    open_ephys_dir = RESOURCES_DIR / "v0.6.x_neuropixels_multiexp_multistream"

    nidaq_rec_prefix = (
        "'recording': OpenEphysBinaryRecordingExtractor: 8 channels - 1 "
        "segments - 30.0kHz - 0.003s, 'experiment_name': 'experiment"
    )
    nidaq_rec_suffix = (
        "'stream_name': 'Record Node 101#NI-DAQmx-103.PXIe-6341'"
    )

    neuropix_rec_prefix = (
        "'recording': OpenEphysBinaryRecordingExtractor: 384 channels - 1 "
        "segments - 30.0kHz - 0.003s, 'experiment_name': 'experiment"
    )
    neuropix_rec_suffix = (
        "'stream_name': 'Record Node 101#Neuropix-PXI-100.Probe"
    )

    expected_read_blocks_info = [
        f"{{{nidaq_rec_prefix}1', {nidaq_rec_suffix}}}",
        f"{{{neuropix_rec_prefix}1', {neuropix_rec_suffix}B'}}",
        f"{{{neuropix_rec_prefix}1', {neuropix_rec_suffix}C'}}",
        f"{{{nidaq_rec_prefix}3', {nidaq_rec_suffix}}}",
        f"{{{neuropix_rec_prefix}3', {neuropix_rec_suffix}B'}}",
        f"{{{neuropix_rec_prefix}3', {neuropix_rec_suffix}C'}}",
        f"{{{nidaq_rec_prefix}6', {nidaq_rec_suffix}}}",
        f"{{{neuropix_rec_prefix}6', {neuropix_rec_suffix}B'}}",
        f"{{{neuropix_rec_prefix}6', {neuropix_rec_suffix}C'}}",
    ]

    def test_read(self):
        openephys_reader = EphysReaders.Readers.openephys.name
        read_blocks = EphysReaders.get_read_blocks(
            openephys_reader, self.open_ephys_dir
        )
        read_blocks_info = [str(r) for r in read_blocks]
        self.assertEqual(self.expected_read_blocks_info, read_blocks_info)

    def test_read_fails(self):
        with self.assertRaises(Exception):
            list(
                EphysReaders.get_read_blocks(
                    "Made up reader", self.open_ephys_dir
                )
            )


class TestImagingReaders(unittest.TestCase):

    imaging_dir = RESOURCES_DIR / "imaging"
    exaspim_dir = imaging_dir / "exaSPIM_125L_2022-08-05_17-25-36"
    mesospim_dir = imaging_dir / "mesoSPIM_125L_2022-08-18_17-05-00"

    @parameterized.expand([
        (ImagingReaders.Readers.exaspim.value, exaspim_dir),
        (ImagingReaders.Readers.mesospim.value, mesospim_dir),
    ])
    def test_get_raw_data_dir(self, reader, input_dir):
        raw_data_dir = ImagingReaders.get_raw_data_dir(reader, input_dir)
        expected_raw_data_dir = input_dir / reader
        self.assertEqual(expected_raw_data_dir, raw_data_dir)

    @parameterized.expand([
        ImagingReaders.Readers.exaspim.value,
        ImagingReaders.Readers.mesospim.value,
    ])
    def test_get_raw_data_dir_fails(self, reader):
        with self.assertRaises(FileNotFoundError):
            ImagingReaders.get_raw_data_dir(reader, "Made up directory")


if __name__ == "__main__":
    unittest.main()



import argparse
import itertools
import logging
import os
import time
from pathlib import PurePath

import numpy as np
from cluster.config import load_jobqueue_config
from dask_jobqueue import SLURMCluster
from distributed import Client
from s3transfer.constants import GB, MB

from aind_data_transfer.s3 import S3Uploader
from aind_data_transfer.util.file_utils import collect_filepaths

LOG_FMT = "%(asctime)s %(message)s"
LOG_DATE_FMT = "%Y-%m-%d %H:%M"

logging.basicConfig(format=LOG_FMT, datefmt=LOG_DATE_FMT)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def chunk_files(input_dir, ntasks, recursive=True, exclude_dirs=None):
    filepaths = collect_filepaths(input_dir, recursive, exclude_dirs)
    logger.info(f"Collected {len(filepaths)} files")
    return np.array_split(filepaths, ntasks)


def upload_files_job(
    input_dir, files, bucket, s3_path, n_threads, target_throughput, part_size, timeout
):
    uploader = S3Uploader(
        num_threads=n_threads,
        target_throughput=target_throughput,
        part_size=part_size,
        upload_timeout=timeout,
    )
    return uploader.upload_files(files, bucket, s3_path, root=input_dir)


def get_client():
    config = load_jobqueue_config()
    slurm_config = config["jobqueue"]["slurm"]
    # cluster config is automatically populated from
    # ~/.config/dask/jobqueue.yaml
    cluster = SLURMCluster()
    cluster.scale(slurm_config["n_workers"])
    logger.info(cluster.job_script())
    client = Client(cluster)
    return client, slurm_config


def run_cluster_job(
    input_dir,
    bucket,
    s3_path,
    target_throughput,
    part_size,
    timeout,
    parallelism,
    recursive,
    exclude_dirs=None
):
    client, config = get_client()
    ntasks = config["n_workers"]
    cores = config["cores"]

    chunked_files = chunk_files(input_dir, ntasks * parallelism, recursive, exclude_dirs)
    logger.info(
        f"Split files into {len(chunked_files)} chunks with "
        f"{len(chunked_files[0])} files each"
    )

    futures = []
    for chunk in chunked_files:
        futures.append(
            client.submit(
                upload_files_job,
                input_dir=input_dir,
                files=chunk,
                bucket=bucket,
                s3_path=s3_path,
                n_threads=cores,
                target_throughput=target_throughput,
                part_size=part_size,
                timeout=timeout,
            )
        )
    failed_uploads = list(itertools.chain(*client.gather(futures)))
    logger.info(f"{len(failed_uploads)} failed uploads:\n{failed_uploads}")
    client.close()


def run_local_job(
    input_dir,
    bucket,
    s3_path,
    nthreads,
    target_throughput,
    part_size,
    timeout,
    recursive,
    exclude_dirs=None
):
    uploader = S3Uploader(
        num_threads=nthreads,
        target_throughput=target_throughput,
        part_size=part_size,
        upload_timeout=timeout,
    )
    if os.path.isdir(input_dir):
        failed_uploads = uploader.upload_folder(
            input_dir, bucket, s3_path, recursive, exclude_dirs
        )
    elif os.path.isfile(input_dir):
        failed_uploads = uploader.upload_file(input_dir, bucket, s3_path)
    else:
        raise ValueError(
            f"Invalid value for --input: {input_dir} does not exist"
        )
    logger.info(f"{len(failed_uploads)} failed uploads:\n{failed_uploads}")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        type=str,
        help="folder or file to upload",
    )
    parser.add_argument(
        "-b",
        "--bucket",
        type=str,
        help="s3 bucket",
    )
    parser.add_argument(
        "--s3_path",
        type=str,
        default=None,
        help='s3 path relative to bucket, e.g., "folder/data.h5"',
    )
    parser.add_argument(
        "--target_throughput",
        type=float,
        default=100 * GB / 8,
        help="target throughput (bytes)",
    )
    parser.add_argument(
        "--part_size",
        type=int,
        default=64 * MB,
        help="part size for s3 multipart uploads (bytes)",
    )
    parser.add_argument(
        "--timeout",
        type=float,
        default=None,
        help="per-file upload timeout (s). Default is None.",
    )
    parser.add_argument(
        "--cluster",
        default=False,
        action="store_true",
        help="run in cluster mode",
    )
    parser.add_argument(
        "--batch_num",
        type=int,
        default=3,
        help="number of tasks per job. "
        "Increase this if you run into worker memory issues",
    )
    parser.add_argument(
        "--nthreads",
        type=int,
        default=1,
        help="num threads to use if running locally",
    )
    parser.add_argument(
        "--recursive",
        default=False,
        action="store_true",
        help="upload a directory recursively",
    )
    parser.add_argument(
        "--exclude_dirs",
        type=str,
        nargs='+',
        default=None,
        help="directories to exclude from upload"
    )

    args = parser.parse_args()

    input_path = args.input
    bucket = args.bucket

    s3_path = args.s3_path
    if s3_path is None:
        s3_path = PurePath(args.input).name
    # remove leading slash since it will result in the structure
    # "bucket/'/'/s3_path" instead of "bucket/s3_path"
    s3_path = s3_path.strip("/")
    logger.info(f"Will upload to {args.bucket}/{s3_path}")

    t0 = time.time()
    if args.cluster:
        run_cluster_job(
            input_dir=input_path,
            bucket=bucket,
            s3_path=s3_path,
            target_throughput=args.target_throughput,
            part_size=args.part_size,
            timeout=args.timeout,
            parallelism=args.batch_num,
            recursive=args.recursive,
            exclude_dirs=args.exclude_dirs
        )
    else:
        run_local_job(
            input_dir=input_path,
            bucket=bucket,
            s3_path=s3_path,
            nthreads=args.nthreads,
            target_throughput=args.target_throughput,
            part_size=args.part_size,
            timeout=args.timeout,
            recursive=args.recursive,
            exclude_dirs=args.exclude_dirs
        )

    logger.info(f"Upload done. Took {time.time() - t0}s ")


if __name__ == "__main__":
    main()



"""Loads Ephys job configurations"""
import argparse
import os
import re
from pathlib import Path

import yaml
from numcodecs import Blosc

from aind_data_transfer.readers import EphysReaders


class EphysJobConfigurationLoader:
    """Class to handle loading ephys job configs"""

    def __remove_none(self, data):
        """Remove keys whose value is None."""
        if isinstance(data, dict):
            return {
                k: self.__remove_none(v)
                for k, v in data.items()
                if v is not None
            }
        else:
            return data

    @staticmethod
    def __parse_compressor_configs(configs):
        """Util method to map a string to class attribute"""
        try:
            compressor_name = configs["compress_data_job"]["compressor"][
                "compressor_name"
            ]
        except KeyError:
            compressor_name = None
        try:
            compressor_kwargs = configs["compress_data_job"]["compressor"][
                "kwargs"
            ]
        except KeyError:
            compressor_kwargs = None
        if (
            compressor_name
            and compressor_name == Blosc.codec_id
            and compressor_kwargs
            and "shuffle" in compressor_kwargs
        ):
            shuffle_str = configs["compress_data_job"]["compressor"]["kwargs"][
                "shuffle"
            ]
            shuffle_val = getattr(Blosc, shuffle_str)
            configs["compress_data_job"]["compressor"]["kwargs"][
                "shuffle"
            ] = shuffle_val

    @staticmethod
    def __resolve_endpoints(configs):
        """
        Only the raw data source needs to be provided as long as the base dir
        name is formatted correctly. If the dest_data_dir and cloud endpoints
        are not set in the conf file, they will be created automatically based
        on the name of the raw_data_source.
        Args:
            configs (dic): Configurations

        Returns:
            None, modifies the base configs in place
        """
        raw_data_folder = Path(configs["endpoints"]["raw_data_dir"]).name

        dest_data_dir = configs["endpoints"]["dest_data_dir"]
        if dest_data_dir is None and re.match(
            EphysReaders.SourceRegexPatterns.subject_datetime.value,
            raw_data_folder,
        ):
            configs["endpoints"]["dest_data_dir"] = (
                "ecephys_" + raw_data_folder
            )
        if dest_data_dir is None and re.match(
            EphysReaders.SourceRegexPatterns.ecephys_subject_datetime.value,
            raw_data_folder,
        ):
            configs["endpoints"]["dest_data_dir"] = raw_data_folder

        if configs["endpoints"]["s3_prefix"] is None:
            dest_data_folder = Path(configs["endpoints"]["dest_data_dir"]).name
            configs["endpoints"]["s3_prefix"] = dest_data_folder

        if configs["endpoints"]["gcp_prefix"] is None:
            dest_data_folder = Path(configs["endpoints"]["dest_data_dir"]).name
            configs["endpoints"]["gcp_prefix"] = dest_data_folder

        if configs["trigger_codeocean_job"]["job_type"] is None:
            configs["trigger_codeocean_job"]["job_type"] = (
                configs["data"]["name"])

        if configs["trigger_codeocean_job"]["bucket"] is None:
            configs["trigger_codeocean_job"]["bucket"] = (
                configs["endpoints"]["s3_bucket"])

        if configs["trigger_codeocean_job"]["prefix"] is None:
            configs["trigger_codeocean_job"]["prefix"] = (
                configs["endpoints"]["s3_prefix"])

    @staticmethod
    def __resolve_logging(configs: dict) -> None:
        """
        Resolves logging config in place
        Parameters
        ----------
        configs : dict
          Configurations

        Returns
        -------
        None
        """

        if configs["logging"]["level"] is None:
            configs["logging"]["level"] = "INFO"
        if os.getenv("LOG_LEVEL"):
            configs["logging"]["level"] = os.getenv("LOG_LEVEL")

    def load_configs(self, sys_args):
        """Load yaml config at conf_src Path as python dict"""
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-c", "--conf-file-location", required=True, type=str
        )
        parser.add_argument(
            "-r", "--raw-data-source", required=False, type=str
        )
        args = parser.parse_args(sys_args)
        conf_src = args.conf_file_location
        with open(conf_src) as f:
            raw_config = yaml.load(f, Loader=yaml.SafeLoader)
        if args.raw_data_source is not None:
            raw_config["endpoints"]["raw_data_dir"] = args.raw_data_source
        self.__resolve_endpoints(raw_config)
        self.__resolve_logging(raw_config)
        config_without_nones = self.__remove_none(raw_config)
        self.__parse_compressor_configs(config_without_nones)
        return config_without_nones



import boto3
from botocore.exceptions import ClientError


def get_secret(secret_name, region_name):
    """Retrieves a secret"""

    # secret_name = "video_encryption_password"
    # region_name = "us-west-2"

    # Create a Secrets Manager client
    session = boto3.session.Session()
    client = session.client(
        service_name="secretsmanager", region_name=region_name
    )

    try:
        get_secret_value_response = client.get_secret_value(
            SecretId=secret_name
        )
    except ClientError as e:
        # For a list of exceptions thrown, see
        # https://docs.aws.amazon.com/secretsmanager/latest/apireference/
        #   API_GetSecretValue.html
        raise e

    # Decrypts secret using the associated KMS key.
    # Don't log or print!
    secret = get_secret_value_response["SecretString"]

    return secret



import fnmatch
import logging
import time
from pathlib import Path
from typing import Union, List

import dask
import numpy as np
import zarr
from aicsimageio.types import PhysicalPixelSizes
from aicsimageio.writers import OmeZarrWriter
from distributed import wait
from numpy.typing import NDArray

from aind_data_transfer.util.chunk_utils import (
    guess_chunks,
    expand_chunks,
    ensure_shape_5d,
    ensure_array_5d,
)
from aind_data_transfer.util.file_utils import collect_filepaths
from aind_data_transfer.util.io_utils import (
    DataReaderFactory,
    ImarisReader,
    MissingDatasetError,
)

logging.basicConfig(format="%(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M")
LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def write_files(
    image_paths: list,
    output: str,
    n_levels: int,
    scale_factor: float,
    overwrite: bool = True,
    chunk_size: float = 64,  # MB
    chunk_shape: tuple = None,
    voxel_size: tuple = None,
    storage_options: dict = None,
) -> list:
    """
    Write each image as a separate group to a Zarr store.
    Each group will be named according to the image filename minus extension.
    For example, if a file has path /data/tiff/tile_0_0.tif, the group name
    for that image is tile_0_0. Within each group there are n_levels arrays,
    one for each resolution level.
    Args:
        image_paths: the list of image filepaths to convert to Zarr
        output: the location of the output Zarr store (filesystem, s3, gs)
        n_levels: number of downsampling levels
        scale_factor: scale factor for downsampling in X, Y and Z
        overwrite: whether to overwrite image groups that already exist
        chunk_size: the target chunk size in MB
        chunk_shape: the chunk shape, if None will be computed from chunk_size
        voxel_size: three element tuple giving physical voxel sizes in Z,Y,X order
        storage_options: a dictionary of options to pass to the Zarr storage backend, e.g., "compressor"
    Returns:
        A list of metrics for each converted image
    """
    if chunk_shape is None and chunk_size <= 0:
        raise ValueError(
            "Either chunk_shape must be set or chunk_size must be greater than 0"
        )
    if chunk_shape is not None and any(s < 1 for s in chunk_shape):
        raise ValueError("chunk_shape must be at least 1 in all dimensions")

    if not image_paths:
        LOGGER.warning("No images found. Exiting.")
        return []

    if storage_options is None:
        storage_options = {}

    writer = OmeZarrWriter(output)

    all_metrics = []

    for impath in image_paths:
        LOGGER.info(f"Writing tile {impath}")

        tile_name = Path(impath).stem

        if not overwrite and _tile_exists(output, tile_name, n_levels):
            LOGGER.info(f"Skipping tile {tile_name}, already exists.")
            continue

        tile_metrics = {
            "tile": Path(impath).absolute(),
        }
        try:
            tile_metrics.update(storage_options["params"])
        except KeyError:
            LOGGER.warning("compression parameters will not be logged")

        # Create reader, but don't construct dask array
        # until we know the optimal chunk shape.
        reader = DataReaderFactory().create(impath)

        # TODO: pass units to zarr writer
        if voxel_size is None:
            try:
                voxel_size, unit = reader.get_voxel_size()
            except Exception:
                voxel_size = [1.0, 1.0, 1.0]
                unit = "um"
        LOGGER.info(f"Using voxel size: {voxel_size}")
        physical_pixel_sizes = PhysicalPixelSizes(
            Z=voxel_size[0], Y=voxel_size[1], X=voxel_size[2]
        )

        # Attempt to parse tile origin
        origin = _parse_origin(reader)

        # We determine the chunk size before creating the dask array since
        # rechunking an existing dask array, e.g, data = data.rechunk(chunks),
        # causes memory use to grow (unbounded?) during the zarr write step.
        # See https://github.com/dask/dask/issues/5105.
        if chunk_shape is None:
            chunks = _compute_chunks(reader, chunk_size)
        else:
            chunks = tuple(chunk_shape)

        LOGGER.info(
            f"chunks: {chunks}, {np.product(chunks) * reader.get_itemsize() / (1024 ** 2)} MiB"
        )

        tile_metrics["chunks"] = chunks

        # Get the chunk dimensions that exist in the original, un-padded image
        reader_chunks = chunks[len(chunks) - len(reader.get_shape()) :]

        pyramid = _get_or_create_pyramid(reader, n_levels, reader_chunks)

        for i in range(len(pyramid)):
            pyramid[i] = ensure_array_5d(pyramid[i])

        LOGGER.info(f"{pyramid[0]}")

        LOGGER.info("Starting write...")
        t0 = time.time()
        jobs = writer.write_multiscale(
            pyramid=pyramid,
            image_name=tile_name,
            physical_pixel_sizes=physical_pixel_sizes,
            translation=origin,
            channel_names=None,
            channel_colors=None,
            scale_factor=(scale_factor,) * 3,
            chunks=chunks,
            storage_options=storage_options,
            compute_dask=False,
        )
        if jobs:
            LOGGER.info("Computing dask arrays...")
            arrs = dask.persist(*jobs)
            wait(arrs)
        write_time = time.time() - t0

        _populate_metrics(
            tile_metrics,
            tile_name,
            output,
            _get_bytes(pyramid),
            write_time,
            n_levels,
            pyramid[0].shape,
            pyramid[0].dtype,
        )

        LOGGER.info(
            f"Finished writing tile {tile_name}.\n"
            f"Took {write_time}s. {tile_metrics['write_bps'] / (1024 ** 2)} MiB/s"
        )

        all_metrics.append(tile_metrics)

        reader.close()

    return all_metrics


def write_folder(
    input: str,
    output: str,
    n_levels: int,
    scale_factor: float,
    overwrite: bool = True,
    chunk_size: float = 64,  # MB
    chunk_shape: tuple = None,
    voxel_size: tuple = None,
    exclude: list = None,
    storage_options: dict = None,
    recursive: bool = False,
) -> list:
    """
    Write each image in the input directory as a separate group
    to a Zarr store. Each group will be named according to the image filename
    minus extension. For example, if a file has path /data/tiff/tile_0_0.tif, the group name
    for that image is tile_0_0. Within each group there are n_levels arrays, one for each resolution level.
    Args:
        input: the directory of images to convert to Zarr
        output: the location of the output Zarr store (filesystem, s3, gs)
        n_levels: number of downsampling levels
        scale_factor: scale factor for downsampling in X, Y and Z
        overwrite: whether to overwrite image groups that already exist
        chunk_size: the target chunk size in MB
        chunk_shape: the chunk shape, if None will be computed from chunk_size
        voxel_size: three element tuple giving physical voxel sizes in Z,Y,X order
        exclude: a list of filename patterns to exclude from conversion
        storage_options: a dictionary of options to pass to the Zarr storage backend, e.g., "compressor"
        recursive: whether to convert all images in all subfolders
    Returns:
        A list of metrics for each converted image
    """
    if exclude is None:
        exclude = []

    image_paths = collect_filepaths(
        input,
        recursive=recursive,
        include_exts=DataReaderFactory().VALID_EXTENSIONS,
    )

    exclude_paths = set()
    for path in image_paths:
        if any(fnmatch.fnmatch(path, pattern) for pattern in exclude):
            exclude_paths.add(path)

    image_paths = [p for p in image_paths if p not in exclude_paths]
    LOGGER.info(f"Found {len(image_paths)} images to process")

    return write_files(
        image_paths,
        output,
        n_levels,
        scale_factor,
        overwrite,
        chunk_size,
        chunk_shape,
        voxel_size,
        storage_options,
    )


def _parse_origin(reader):
    origin = None
    try:
        origin = reader.get_origin()
        while len(origin) < 5:
            origin = (0, *origin)
    except AttributeError:
        LOGGER.warning("Origin metadata could not be parsed from image")
    LOGGER.info(f"Using origin: {origin}")
    return origin


def _tile_exists(zarr_path, tile_name, n_levels):
    z = zarr.open(zarr_path, "r")
    try:
        # TODO: only re-upload missing levels
        a = z[f"{tile_name}/{n_levels - 1}"]
        return a.nbytes > 0
    except KeyError:
        return False


def _compute_chunks(reader, target_size_mb):
    target_size_bytes = target_size_mb * 1024 * 1024
    padded_chunks = ensure_shape_5d(reader.get_chunks())
    padded_shape = ensure_shape_5d(reader.get_shape())
    spatial_chunks = padded_chunks[2:]
    spatial_shape = padded_shape[2:]
    LOGGER.info(f"Using multiple of base chunk size: {padded_chunks}")
    if spatial_chunks[1:] == spatial_shape[1:]:
        LOGGER.info(
            "chunks and shape have same XY dimensions, "
            "will chunk along Z only."
        )
        chunks = guess_chunks(
            spatial_shape, target_size_bytes, reader.get_itemsize(), mode="z"
        )
    else:
        chunks = expand_chunks(
            spatial_chunks,
            spatial_shape,
            target_size_bytes,
            reader.get_itemsize(),
            mode="cycle",
        )
    chunks = ensure_shape_5d(chunks)
    return chunks


def _create_pyramid(data, n_lvls):
    from xarray_multiscale import multiscale
    from xarray_multiscale.reducers import windowed_mean

    pyramid = multiscale(
        data,
        windowed_mean,  # func
        (2,) * data.ndim,  # scale factors
        depth=n_lvls - 1,
        preserve_dtype=True,
    )
    return [arr.data for arr in pyramid]


def _get_or_create_pyramid(reader, n_levels, chunks):
    if isinstance(reader, ImarisReader):
        try:
            pyramid = reader.get_dask_pyramid(
                n_levels,
                timepoint=0,
                channel=0,
                # Use only the dimensions that exist in the base image.
                chunks=chunks,
            )
        except MissingDatasetError as e:
            LOGGER.error(e)
            LOGGER.warning(
                f"{reader.get_filepath()} does not contain all requested scales."
                f"Computing them instead..."
            )
            pyramid = _create_pyramid(
                reader.as_dask_array(chunks=chunks), n_levels
            )
    else:
        pyramid = _create_pyramid(
            reader.as_dask_array(chunks=chunks), n_levels
        )

    return pyramid


def _get_bytes(data: Union[List[NDArray], NDArray]):
    if isinstance(data, list):
        total_bytes = 0
        for arr in data:
            total_bytes += arr.nbytes
        return total_bytes
    return data.nbytes


def _get_storage_ratio(zarr_path: str, dataset_name: str):
    z = zarr.open(zarr_path, "r")
    full_res = z[f"{dataset_name}/0"]
    return full_res.nbytes / full_res.nbytes_stored


def _get_bytes_stored(zarr_path: str, dataset_name: str, n_levels: int):
    z = zarr.open(zarr_path, "r")
    total_bytes_stored = 0
    for res in range(n_levels):
        arr = z[f"{dataset_name}/{res}"]
        total_bytes_stored += arr.nbytes_stored
    return total_bytes_stored


def _populate_metrics(
    tile_metrics,
    tile_name,
    out_zarr,
    bytes_read,
    write_time,
    n_levels,
    shape,
    dtype,
):
    tile_metrics["n_levels"] = n_levels
    tile_metrics["shape"] = shape
    tile_metrics["dtype"] = dtype
    tile_metrics["write_time"] = write_time
    tile_metrics["bytes_read"] = bytes_read
    tile_metrics["write_bps"] = tile_metrics["bytes_read"] / write_time
    tile_metrics["bytes_stored"] = _get_bytes_stored(
        out_zarr, tile_name, n_levels
    )
    storage_ratio = _get_storage_ratio(out_zarr, tile_name)
    tile_metrics["storage_ratio"] = storage_ratio
    LOGGER.info(f"Compress ratio: {storage_ratio}")



"""Module that contains the API to retrieve a Compressor for Ephys Data.
"""
import os
from enum import Enum
from pathlib import Path
from typing import Optional

import numpy as np
import pyminizip
import spikeinterface.full as si
import spikeinterface.preprocessing as spre
from numcodecs import Blosc
from tqdm import tqdm
from wavpack_numcodecs import WavPack

from aind_data_transfer.readers import EphysReaders


class EphysCompressors:
    """This class contains the methods to retrieve a compressor, and to scale
    a read block by lsb and median values.
    """

    class Compressors(Enum):
        """Enum for compression algorithms a user can select"""

        blosc = Blosc.codec_id
        wavpack = "wavpack"

    compressors = [member.value for member in Compressors]

    @staticmethod
    def get_compressor(compressor_name, **kwargs):
        """
        Retrieve a compressor for a given name and optional kwargs.
        Args:
            compressor_name (str): Matches one of the names Compressors enum
            **kwargs (dict): Options to pass into the Compressor
        Returns:
            An instantiated compressor class.
        """
        if compressor_name == EphysCompressors.Compressors.blosc.name:
            return Blosc(**kwargs)
        elif compressor_name == EphysCompressors.Compressors.wavpack.name:
            return WavPack(**kwargs)
        else:
            raise Exception(
                f"Unknown compressor. Please select one of "
                f"{EphysCompressors.compressors}"
            )

    @staticmethod
    def _get_median_and_lsb(
        recording,
        disable_tqdm=False,
        **random_chunk_kwargs,
    ):
        """This function estimates the channel-wise medians and the overall
        LSB from a recording
        Parameters
        ----------
        recording : si.BaseRecording
            The input recording object
        disable_tqdm : bool, optional
            Disable progress bar, default is False
        **random_chunk_kwargs: keyword arguments for
            si.get_random_data_chunks() (chunk_size, num_chunks_per_segment)
        Returns
        -------
        int
            lsb_value
        np.array
            median_values
        """
        # gather chunks
        chunks = si.get_random_data_chunks(
            recording, seed=0, **random_chunk_kwargs
        )

        lsb_value = 0
        num_channels = recording.get_num_channels()
        dtype = recording.get_dtype()

        channel_idxs = np.arange(num_channels)
        min_values = np.zeros(num_channels, dtype=dtype)
        median_values = np.zeros(num_channels, dtype=dtype)
        offsets = np.zeros(num_channels, dtype=dtype)

        for ch in tqdm(
            channel_idxs, desc="Estimating channel stats", disable=disable_tqdm
        ):
            unique_vals = np.unique(chunks[:, ch])
            unique_vals_abs = np.abs(unique_vals)
            # it might happen that there is a single unique value
            # (e.g. a channel is broken)
            if len(unique_vals) > 1:
                lsb_val = np.min(np.diff(unique_vals))
            else:
                # in this case we can't estimate the LSB for the channel
                lsb_val = 0

            min_values[ch] = np.min(unique_vals_abs)
            median_values[ch] = np.median(chunks[:, ch]).astype(dtype)

            unique_vals_m = np.unique(chunks[:, ch] - median_values[ch])
            unique_vals_abs_m = np.abs(unique_vals_m)
            offsets[ch] = np.min(unique_vals_abs_m)

            if lsb_val > lsb_value:
                lsb_value = lsb_val

        return lsb_value, median_values

    @staticmethod
    def scale_read_blocks(
        read_blocks,
        num_chunks_per_segment=100,
        chunk_size=10000,
        disable_tqdm=False,
    ):
        """
        Scales a read_block. A read_block is dict of
        {'recording', 'block_index', 'stream_name'}.
        Args:
            read_blocks (iterable): A generator of read_blocks
            num_chunks_per_segment (int):
            chunk_size (int):
            disable_tqdm (boolean): Optionally disable a progress bar.
              Defaults to False.
        Returns:
            A generated scaled_read_block. A dict of
            {'scaled_recording', 'block_index', 'stream_name'}.

        """
        for read_block in read_blocks:
            # We don't need to scale the NI-DAQ recordings
            # TODO: Convert this to regex matching?
            if (
                EphysReaders.RecordingBlockPrefixes.nidaq.value
                in read_block["stream_name"]
            ):
                rec_to_compress = read_block["recording"]
            else:
                (
                    lsb_value,
                    median_values,
                ) = EphysCompressors._get_median_and_lsb(
                    read_block["recording"],
                    num_chunks_per_segment=num_chunks_per_segment,
                    chunk_size=chunk_size,
                    disable_tqdm=disable_tqdm,
                )
                dtype = read_block["recording"].get_dtype()
                rec_to_compress = spre.scale(
                    read_block["recording"],
                    gain=1.0,
                    offset=-median_values,
                    dtype=dtype,
                )
                rec_to_compress = spre.scale(
                    rec_to_compress, gain=1.0 / lsb_value, dtype=dtype
                )
                rec_to_compress.set_channel_gains(
                    rec_to_compress.get_channel_gains() * lsb_value
                )
            yield (
                {
                    "scaled_recording": rec_to_compress,
                    "experiment_name": read_block["experiment_name"],
                    "stream_name": read_block["stream_name"],
                }
            )


class ImagingCompressors:
    class Compressors(Enum):
        """Enum for compression algorithms a user can select"""

        blosc = Blosc.codec_id

    compressors = [member.value for member in Compressors]

    @staticmethod
    def get_compressor(compressor_name, **kwargs):
        """
        Retrieve a compressor for a given name and optional kwargs.
        Args:
            compressor_name (str): Matches one of the names Compressors enum
            **kwargs (dict): Options to pass into the Compressor
        Returns:
            An instantiated compressor class.
        """
        if compressor_name == ImagingCompressors.Compressors.blosc.name:
            return Blosc(**kwargs)
        else:
            raise Exception(
                f"Unknown compressor. Please select one of "
                f"{ImagingCompressors.compressors}"
            )


class VideoCompressor:
    """Class to handle video compression and encryption."""

    def __init__(
        self, compression_level: int = 5, encryption_key: Optional[str] = None
    ) -> None:
        """
        Creates a video compressor with compression level and encryption key
        Parameters
        ----------
        compression_level : int
          Integer between 1 and 9. Default is 5.
        encryption_key : Optional[str]
          Optional password to use. Default is None.
        """
        self.compression_level = compression_level
        self.encryption_key = encryption_key

    class VideoFileTypes(Enum):
        """Enum for types of video to compress"""

        MPEG_4 = ".mp4"
        QUICKTIME_MOVIE = ".mov"
        WINDOWSMEDIA_VIEWER = ".wmv"
        AUDIO_VIDEO_INTERLEAVE = ".avi"

    video_file_extensions = tuple([member.value for member in VideoFileTypes])

    def compress_all_videos_in_dir(self, video_dir: Path) -> None:
        """
        Compress and optionally encrypt video files in a directory
        Parameters
        ----------
        video_dir : Path
          Directory where video files are stored.

        Returns
        -------
        None

        """

        for root, dirs, files in os.walk(video_dir):
            for file in files:
                if file.endswith(self.video_file_extensions):
                    raw_file_path = os.path.join(root, file)
                    zip_file_path = os.path.join(root, file) + ".zip"
                    pyminizip.compress(
                        str(raw_file_path),
                        None,
                        str(zip_file_path),
                        self.encryption_key,
                        self.compression_level,
                    )
                    os.remove(raw_file_path)



"""Tests configurations are set properly"""
import os
import unittest
from pathlib import Path

from numcodecs import Blosc

from aind_data_transfer.config_loader.configuration_loader import (
    ImagingJobConfigurationLoader,
)
from aind_data_transfer.config_loader.ephys_configuration_loader import (
    EphysJobConfigurationLoader,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
CONFIGS_DIR = TEST_DIR / "resources" / "test_configs"


class TestEphysJobConfigs(unittest.TestCase):
    """Tests ephys job pipeline methods"""

    def test_conf_loads(self):
        """Basic config loads test"""

        raw_data_dir = (
            "tests/resources/v0.6.x_neuropixels_multiexp_multistream"
        )
        dest_data_dir = (
            "tests/resources/new/v0.6.x_neuropixels_multiexp_multistream"
        )

        expected_configs = {
            "jobs": {
                "clip": True,
                "compress": True,
                "attach_metadata": True,
                "upload_to_s3": True,
                "upload_to_gcp": True,
                "trigger_codeocean_job": False,
            },
            "endpoints": {
                "raw_data_dir": raw_data_dir,
                "dest_data_dir": dest_data_dir,
                "s3_bucket": "some-s3-bucket",
                "s3_prefix": "v0.6.x_neuropixels_multiexp_multistream",
                "gcp_bucket": "some-gcp-bucket",
                "gcp_prefix": "test_20221001",
                "codeocean_domain": "https://acmecorp.codeocean.com",
                "code_repo_location": "https://location_of_code_repo",
            },
            "aws_secret_names": {
                "code_ocean_api_token_name": "secret_name_for_api_token",
                "region": "us-west-2"
            },
            "data": {"name": "openephys"},
            "clip_data_job": {
                "clip_kwargs": {},
            },
            "compress_data_job": {
                "write_kwargs": {
                    "n_jobs": -1,
                    "chunk_duration": "1s",
                    "progress_bar": True,
                },
                "format_kwargs": {},
                "compressor": {
                    "compressor_name": "blosc",
                    "kwargs": {"shuffle": Blosc.BITSHUFFLE},
                },
                "scale_params": {"chunk_size": 20},
            },
            "upload_data_job": {"dryrun": True},
            "trigger_codeocean_job": {
                "job_type": "openephys",
                "bucket": "some-s3-bucket",
                "prefix": "v0.6.x_neuropixels_multiexp_multistream",
            },
            "logging": {"level": "INFO"},
        }
        conf_file_path = CONFIGS_DIR / "ephys_upload_job_test_configs.yml"

        args = ["-c", str(conf_file_path)]

        loaded_configs = EphysJobConfigurationLoader().load_configs(args)
        self.assertEqual(loaded_configs, expected_configs)

    def test_endpoints_are_resolved(self):
        """Tests default endpoints are resolved correctly"""

        raw_data_dir = "/some/random/folder/625463_2022-10-06_10-14-25"
        expected_configs = {
            "jobs": {
                "clip": True,
                "compress": True,
                "attach_metadata": True,
                "upload_to_s3": True,
                "upload_to_gcp": True,
                "trigger_codeocean_job": False,
            },
            "endpoints": {
                "raw_data_dir": raw_data_dir,
                "dest_data_dir": "ecephys_625463_2022-10-06_10-14-25",
                "s3_bucket": "some-s3-bucket",
                "s3_prefix": "ecephys_625463_2022-10-06_10-14-25",
                "gcp_bucket": "some-gcp-bucket",
                "gcp_prefix": "ecephys_625463_2022-10-06_10-14-25",
                "codeocean_domain": "https://acmecorp.codeocean.com",
                "code_repo_location": "https://location_of_code_repo",
            },
            "aws_secret_names": {
                "code_ocean_api_token_name": "secret_name_for_api_token",
                "region": "us-west-2",
                "video_encryption_password": "secret_name_for_vid_password",
            },
            "data": {"name": "openephys"},
            "clip_data_job": {"clip_kwargs": {}},
            "compress_data_job": {
                "write_kwargs": {
                    "n_jobs": -1,
                    "chunk_duration": "1s",
                    "progress_bar": True,
                },
                "format_kwargs": {},
                "compressor": {
                    "compressor_name": "wavpack",
                    "kwargs": {"level": 3},
                },
                "scale_params": {},
            },
            "upload_data_job": {"dryrun": True},
            "trigger_codeocean_job": {
                "job_type": "openephys",
                "bucket": "some-s3-bucket",
                "prefix": "ecephys_625463_2022-10-06_10-14-25",
            },
            "logging": {"level": "INFO"},
        }

        conf_file_path1 = CONFIGS_DIR / "example_configs_src_pattern1.yml"
        conf_file_path2 = CONFIGS_DIR / "example_configs_src_pattern2.yml"
        args1 = ["-c", str(conf_file_path1)]
        args2 = ["-c", str(conf_file_path2)]
        loaded_configs1 = EphysJobConfigurationLoader().load_configs(args1)
        loaded_configs2 = EphysJobConfigurationLoader().load_configs(args2)
        self.assertEqual(loaded_configs1, expected_configs)
        self.assertEqual(loaded_configs2, expected_configs)


class TestImagingJobConfigs(unittest.TestCase):

    code_repo_url = "https://github.com/AllenNeuralDynamics/aind-data-transfer"

    """Basic config loads test"""

    def test_conf_loads(self):
        """Basic config loads test"""

        raw_data_dir = (
            "tests/resources/imaging/exaSPIM_125L_2022-08-05_17-25-36"
        )

        expected_configs = {
            "jobs": {
                "upload_aux_files": False,
                "transcode": True,
            },
            "endpoints": {
                "raw_data_dir": raw_data_dir,
                "dest_data_dir": (
                    "s3://aind-transfer-test/"
                    "exaSPIM_125L_2022-08-05_17-25-36"
                ),
                "code_repo_location": self.code_repo_url,
            },
            "data": {"name": "imaging"},
            "transcode_job": {
                "compressor": {
                    "compressor_name": "blosc",
                    "kwargs": {
                        "cname": "zstd",
                        "clevel": 1,
                        "shuffle": Blosc.SHUFFLE,
                    },
                },
                "chunk_size": 64,
                "resume": False,
                "n_levels": 8,
                "submit_args": {
                    "conda_activate": (
                        "/allen/programs/aind/workgroups/msma/"
                        "cameron.arshadi/miniconda3/bin/"
                        "activate"
                    ),
                    "conda_env": "aind-data-transfer",
                    "cpus_per_task": 1,
                    "mail_user": "cameron.arshadi@alleninstitute.org",
                    "mem_per_cpu": 3000,
                    "nodes": 8,
                    "ntasks_per_node": 8,
                    "run_parent_dir": (
                        "/home/cameron.arshadi/"
                        "exaSPIM-transcode-jobs/"
                        "exaSPIM_125L_2022-08-05_17-25-36"
                    ),
                    "tmp_space": "8GB",
                    "walltime": "72:00:00",
                    "queue": "aind",
                },
            },
        }
        conf_file_path = (
            CONFIGS_DIR / "imaging" / "transcode_job_test_config.yml"
        )

        args = ["-c", str(conf_file_path)]

        loaded_configs = ImagingJobConfigurationLoader().load_configs(args)
        self.assertEqual(loaded_configs, expected_configs)


if __name__ == "__main__":
    unittest.main()



"""Module to test that Processing metadata is processed correctly."""

import datetime
import json
import os
import unittest
from pathlib import Path

from aind_data_schema import Processing

from aind_data_transfer.config_loader.ephys_configuration_loader import (
    EphysJobConfigurationLoader,
)
from aind_data_transfer.transformations.metadata_creation import (
    ProcessingMetadata,
)

TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
CONFIGS_DIR = TEST_DIR / "resources" / "test_configs"
METADATA_DIR = TEST_DIR / "resources" / "test_metadata"

with open(METADATA_DIR / "processing.json", "r") as f:
    expected_processing_instance_json = json.load(f)


class TestProcessingMetadata(unittest.TestCase):
    """Tests methods in ProcessingMetadata class"""

    conf_file_path = CONFIGS_DIR / "ephys_upload_job_test_configs.yml"
    args = ["-c", str(conf_file_path)]
    loaded_configs = EphysJobConfigurationLoader().load_configs(args)

    def test_create_processing_metadata(self) -> None:
        """
        Tests that the processing metadata is created correctly.

        Returns:

        """

        start_date_time = datetime.datetime.fromisoformat(
            "2020-10-20T00:00:00.000+00:00"
        )
        end_date_time = datetime.datetime.fromisoformat(
            "2020-10-20T01:00:00.000+00:00"
        )
        input_location = "some_input_location"
        output_location = "some_output_location"
        code_url = "https://github.com/AllenNeuralDynamics/aind-data-transfer"

        parameters = self.loaded_configs

        processing_instance = ProcessingMetadata.ephys_job_to_processing(
            start_date_time=start_date_time,
            end_date_time=end_date_time,
            input_location=input_location,
            output_location=output_location,
            code_url=code_url,
            parameters=parameters,
        )

        # Hack to get match version to be the same as in the example file
        expected_processing_instance_json["data_processes"][0][
            "version"
        ] = processing_instance.data_processes[0].version

        expected_processing_instance = Processing.parse_obj(
            expected_processing_instance_json
        )

        self.assertEqual(expected_processing_instance, processing_instance)


if __name__ == "__main__":
    unittest.main()



from typing import Any

import os
import zarr
from abc import ABC, abstractmethod

import dask.array as da
import dask_image.imread
import h5py
# Importing this alone doesn't work on HPC
# Must manually override HDF5_PLUGIN_PATH environment variable
# in each Dask worker
import hdf5plugin
import numpy as np
import tifffile


class DataReader(ABC):
    def __init__(self, filepath):
        self.filepath = filepath
        super().__init__()

    @abstractmethod
    def as_dask_array(self, chunks: Any = True) -> da.Array:
        pass

    @abstractmethod
    def as_array(self) -> np.ndarray:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def get_shape(self) -> tuple:
        pass

    @abstractmethod
    def get_chunks(self) -> tuple:
        pass

    @abstractmethod
    def get_itemsize(self) -> int:
        pass

    def get_filepath(self) -> str:
        return self.filepath


class TiffReader(DataReader):
    def __init__(self, filepath):
        super().__init__(filepath)
        self.handle = tifffile.TiffFile(filepath)

    def as_dask_array(self, chunks=True):
        # Note, this will only read the first plane of
        # ImageJ-formatted Tiffs that are larger than 4GB
        return dask_image.imread.imread(self.filepath)

    def as_array(self):
        return self.as_zarr()[:]

    def as_zarr(self):
        return zarr.open(self.handle.aszarr(), "r")

    def get_shape(self):
        # Open as Zarr store in case we're dealing with
        # ImageJ hyperstacks
        return self.as_zarr().shape

    def get_chunks(self):
        return self.as_zarr().chunks

    def get_itemsize(self):
        with tifffile.TiffFile(self.filepath) as tif:
            return tif.series[0].dtype.itemsize

    def get_handle(self):
        return self.handle

    def close(self):
        if self.handle is not None:
            self.handle.close()
            self.handle = None


class MissingDatasetError(Exception):
    pass


class ImarisReader(DataReader):
    DEFAULT_DATA_PATH = "/DataSet/ResolutionLevel 0/TimePoint 0/Channel 0/Data"

    def __init__(self, filepath):
        super().__init__(filepath)
        self.handle = h5py.File(self.filepath, mode="r")

    def as_dask_array(self, data_path=DEFAULT_DATA_PATH, chunks=True):
        return da.from_array(self.get_dataset(data_path), chunks=chunks)

    def as_array(self, data_path=DEFAULT_DATA_PATH):
        return self.get_dataset(data_path)[:]

    def get_dataset(self, data_path=DEFAULT_DATA_PATH) -> h5py.Dataset:
        return self.handle[data_path]

    def get_shape(self, data_path=DEFAULT_DATA_PATH):
        return self.get_dataset(data_path).shape

    def get_chunks(self, data_path=DEFAULT_DATA_PATH):
        return self.get_dataset(data_path).chunks

    def get_itemsize(self, data_path=DEFAULT_DATA_PATH):
        return self.get_dataset(data_path).dtype.itemsize

    def get_handle(self):
        return self.handle

    def get_dask_pyramid(
        self, num_levels, timepoint=0, channel=0, chunks: Any = True
    ):
        darrays = []
        for lvl in range(0, num_levels):
            ds_path = f"/DataSet/ResolutionLevel {lvl}/TimePoint {timepoint}/Channel {channel}/Data"
            if ds_path not in self.get_handle():
                raise MissingDatasetError(f"{ds_path} does not exist")

            if isinstance(chunks, bool) or chunks is None:
                lvl_chunks = chunks
            else:
                lvl_shape = self.get_shape(ds_path)
                assert len(chunks) == len(lvl_shape)
                lvl_chunks = list(chunks)
                for i in range(len(chunks)):
                    lvl_chunks[i] = min(chunks[i], lvl_shape[i])

            darrays.append(self.as_dask_array(ds_path, chunks=lvl_chunks))

        return darrays

    def get_dataset_info(self):
        return self.get_handle()[f"DataSetInfo/Image"]

    def get_origin(self):
        info = self.get_dataset_info()
        x_min = float(info.attrs["ExtMin0"].tostring())
        y_min = float(info.attrs["ExtMin1"].tostring())
        z_min = float(info.attrs["ExtMin2"].tostring())
        return [z_min, y_min, x_min]

    def get_extent(self):
        info = self.get_dataset_info()
        x_max = float(info.attrs["ExtMax0"].tostring())
        y_max = float(info.attrs["ExtMax1"].tostring())
        z_max = float(info.attrs["ExtMax2"].tostring())
        return [z_max, y_max, x_max]

    def get_voxel_size(self):
        info = self.get_dataset_info()
        extmin = self.get_origin()
        extmax = self.get_extent()
        x = int(info.attrs["X"].tostring())
        y = int(info.attrs["Y"].tostring())
        z = int(info.attrs["Z"].tostring())
        unit = info.attrs["Unit"].tostring()
        voxsize = [
            (extmax[0] - extmin[0]) / z,
            (extmax[1] - extmin[1]) / y,
            (extmax[2] - extmin[2]) / x,
        ]
        return voxsize, unit

    def close(self):
        if self.handle is not None:
            self.handle.close()
            self.handle = None


class DataReaderFactory:
    VALID_EXTENSIONS = [".tiff", ".tif", ".h5", ".ims"]

    factory = {}

    def __init__(self):
        self.factory[".tif"] = TiffReader
        self.factory[".tiff"] = TiffReader
        self.factory[".h5"] = ImarisReader
        self.factory[".ims"] = ImarisReader

    def get_valid_extensions(self):
        return self.VALID_EXTENSIONS

    def create(self, filepath) -> DataReader:
        _, ext = os.path.splitext(filepath)
        if ext not in self.VALID_EXTENSIONS:
            raise NotImplementedError(f"File type {ext} not supported")
        return self.factory[ext](filepath)



"""This module contains the api to retrieve a reader for ephys data.
"""
import re
from enum import Enum
from pathlib import Path

import numpy as np
import spikeinterface.extractors as se


class EphysReaders:
    """This class contains the methods to retrieve a reader."""

    class Readers(Enum):
        """Enum for readers."""

        openephys = "openephys"

    readers = [member.value for member in Readers]

    class RecordingBlockPrefixes(Enum):
        """Enum for types of recording blocks."""

        # TODO: Convert these to regex patterns?
        neuropix = "Neuropix"
        nidaq = "NI-DAQ"

    class SourceRegexPatterns(Enum):
        """Enum for regex patterns the source folder name should match"""

        subject_datetime = r"\d+_\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}"
        ecephys_subject_datetime = (
            r"ecephys_\d+_\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}"
        )

    @staticmethod
    def get_read_blocks(reader_name, input_dir):
        """
        Given a reader name and input directory, generate a stream of
        recording blocks.
        Args:
            reader_name (str): Matches one of the names Readers enum
            input_dir (Path): Directory of ephys data to read.
        Returns:
            A generator of read blocks. A read_block is dict of
            {'recording', 'experiment_name', 'stream_name'}.

        """
        if reader_name == EphysReaders.Readers.openephys.value:
            nblocks = se.get_neo_num_blocks(reader_name, input_dir)
            stream_names, stream_ids = se.get_neo_streams(
                reader_name, input_dir
            )
            # load first stream to map block_indices to experiment_names
            rec_test = se.read_openephys(
                input_dir, block_index=0, stream_name=stream_names[0]
            )
            record_node = list(rec_test.neo_reader.folder_structure.keys())[0]
            experiments = rec_test.neo_reader.folder_structure[record_node][
                "experiments"
            ]
            exp_ids = list(experiments.keys())
            experiment_names = [
                experiments[exp_id]["name"] for exp_id in sorted(exp_ids)
            ]
            for block_index in range(nblocks):
                for stream_name in stream_names:
                    rec = se.read_openephys(
                        input_dir,
                        stream_name=stream_name,
                        block_index=block_index,
                    )
                    yield (
                        {
                            "recording": rec,
                            "experiment_name": experiment_names[block_index],
                            "stream_name": stream_name,
                        }
                    )
        else:
            raise Exception(
                f"Unknown reader: {reader_name}. "
                f"Please select one of {EphysReaders.readers}"
            )

    @staticmethod
    def get_streams_to_clip(reader_name, input_dir):
        stream_names, stream_ids = se.get_neo_streams(reader_name, input_dir)
        for dat_file in input_dir.glob("**/*.dat"):
            oe_stream_name = dat_file.parent.name
            si_stream_name = [
                stream_name
                for stream_name in stream_names
                if oe_stream_name in stream_name
            ][0]
            n_chan = se.read_openephys(
                input_dir, block_index=0, stream_name=si_stream_name
            ).get_num_channels()
            data = np.memmap(
                dat_file, dtype="int16", order="C", mode="r"
            ).reshape(-1, n_chan)
            yield {
                "data": data,
                "relative_path_name": str(dat_file.relative_to(input_dir)),
                "n_chan": n_chan,
            }


class ImagingReaders:
    """This class contains the methods to retrieve a reader for aind imaging data."""

    class Readers(Enum):
        """Enum for readers."""

        exaspim = "exaSPIM"
        mesospim = "mesoSPIM"

    readers = [member.value for member in Readers]

    class SourceRegexPatterns(Enum):
        """Enum for regex patterns the source folder name should match"""

        exaspim_acquisition = r"exaSPIM_[A-Z0-9]+_\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}"
        mesospim_acquisition = r"mesoSPIM_[A-Z0-9]+_\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}"

    @staticmethod
    def get_raw_data_dir(reader_name, input_dir):
        if reader_name not in ImagingReaders.readers:
            raise Exception(
                f"Unknown reader: {reader_name}. "
                f"Please select one of {ImagingReaders.readers}"
            )
        raw_data_dir = Path(input_dir) / reader_name
        if not raw_data_dir.is_dir():
            raise FileNotFoundError(f"Raw data directory not found: {raw_data_dir}")
        return raw_data_dir

    @staticmethod
    def get_reader_name(input_dir):
        # re.search does not work with Path objects
        input_dir = str(input_dir)
        if re.search(ImagingReaders.Readers.exaspim.value, input_dir) is not None:
            return ImagingReaders.Readers.exaspim.value
        elif re.search(ImagingReaders.Readers.mesospim.value, input_dir) is not None:
            return ImagingReaders.Readers.mesospim.value
        else:
            raise Exception(f"An appropriate reader could not be created for {input_dir}")




"""This module will have classes that handle mapping to metadata files."""

from datetime import datetime

import aind_data_schema
from aind_data_schema.processing import Processing, ProcessName, DataProcess

import aind_data_transfer


class ProcessingMetadata:
    """Class to handle the creation of the processing metadata file."""

    # TODO: import this from aind_data_schema.Processing class
    output_file_name = "processing.json"

    @staticmethod
    def ephys_job_to_processing(
        start_date_time: datetime,
        end_date_time: datetime,
        input_location: str,
        output_location: str,
        code_url: str,
        parameters: dict,
        notes: str = None,
    ) -> aind_data_schema.Processing:
        """
        Creates a processing instance based on the openephys_job settings
        Parameters
        ----------
        start_date_time : datetime
          Time the processing job started
        end_date_time : datetime
          Time the processing job ended
        input_location : str
          Location of the raw data source
        output_location : str
          Location of the processed data
        code_url : str
          Location of where the code is maintained.
          TODO: Extract this from pyproject.toml file.
        parameters : dict
          Parameters of the processing job
        notes : Optional[str]
            Optional notes to attach
        Returns
        -------
        aind_data_schema.Processing
          A Processing instance to annotate a dataset with.

        """
        data_processing_instance = DataProcess(
            name=ProcessName.EPHYS_PREPROCESSING.value,
            version=aind_data_transfer.__version__,
            start_date_time=start_date_time,
            end_date_time=end_date_time,
            input_location=input_location,
            output_location=output_location,
            code_url=code_url,
            parameters=parameters,
            notes=notes,
        )
        processing_instance = Processing(
            data_processes=[data_processing_instance]
        )

        return processing_instance
